/*
 *  DialogButtonBar in org.jpws.front.util
 *  file: DialogButtonBar.java
 * 
 *  Project Jpws-Front
 *  @author Wolfgang Keller
 *  Created 28.09.2004
 *  Version
 * 
 *  Copyright (c) 2005 by Wolfgang Keller, Munich, Germany
 */
package JRunMC.util;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JDialog;


/**
 * Creates a {@link ButtonBar} containing minimal an "Ok" button.
 * Depending to init specification, additional elements are "Cancel" and "Help"
 * buttons.
 * <p>The application must setup a <code>ButtonBarListener</code> in order to
 * properly react to button actions performed by the user.  
 * 
 * @author Wolfgang Keller
 */
public class DialogButtonBar extends ButtonBar
                             implements ActionListener
{

   public static final long serialVersionUID = 113;

   public static final int OK_BUTTON = 0;
   public static final int OK_CANCEL_BUTTON = 2;
   public static final int OK_CANCEL_HELP_BUTTON = 3;
   
   private int          barType;
   private ButtonBarListener listener;
   
   private JButton		okButton;
	private JButton	   cancelButton;
   private JButton      helpButton;

   private boolean		okPressed;
   private boolean      cancelPressed;

	/**
	 * 
	 */
	public DialogButtonBar ( int type )
	{
      barType = type;

      // create buttons according to parameter bar type
      okButton		= new JButton(ResourceLoader.getDisplay("button.ok"));
      okButton.addActionListener( this );
      super.add( okButton );

      if ( type >= OK_CANCEL_BUTTON )
      {
		   cancelButton = new JButton(ResourceLoader.getDisplay("button.cancel"));
         cancelButton.addActionListener( this );
         super.add( cancelButton );
      }

      if ( type >= OK_CANCEL_HELP_BUTTON )
      {
         helpButton = new JButton(ResourceLoader.getDisplay("button.help"));
         helpButton.addActionListener( this );
         super.add( helpButton );
      }
   }
   
   /** Should be called to make this object prone to garbage collection.
    *  This also releases any set <code>ButtonBarListener</code>. */ 
   public void destruct ()
   {
      listener = null;
      if ( cancelButton != null )
      {
         okButton.removeActionListener( this );
         okButton = null;
      }
      if ( cancelButton != null )
      {
         cancelButton.removeActionListener( this );
         cancelButton = null;
      }
      if ( helpButton != null )
      {
         helpButton.removeActionListener( this );
         helpButton = null;
      }
   }  // destruct

   private void performButton ( Object button )
   {
      okPressed   = false;
      cancelPressed = false;
      
      if ( button == null )
         return;
      
      if ( button == okButton )
      {
         okPressed = true;
         if ( listener != null )
         {
            okButton.setEnabled( false );
            listener.okButtonPerformed();
            if ( okButton != null )
               okButton.setEnabled( true );
         }
         else
            disposeDialog();
      }
      else if ( button == cancelButton )
      {
         cancelPressed = true;
         if ( listener != null )
         {
            cancelButton.setEnabled( false );
            listener.cancelButtonPerformed();
            if ( cancelButton != null )
               cancelButton.setEnabled( true );
         }
         else
            disposeDialog();
      }
      if ( button == helpButton )
      {
         if ( listener != null )
         {
            helpButton.setEnabled( false );
            listener.helpButtonPerformed();
            if ( helpButton != null )
               helpButton.setEnabled( true );
         }
      }
   }  // performButton

   private void disposeDialog ()
   {
      Component c;
      
      c = this;
      while ( (c=c.getParent()) != null && !(c instanceof JDialog) );
      try { ((JDialog)c).dispose(); }
      catch ( Exception e )
      {}
   }
   
   /** Used to implement the internal button action listener. Does nothing for 
    *  external callers. 
    */
   public void actionPerformed( final ActionEvent e )
   {
      Thread thread;
      
      if ( listener != null )
      {
         thread = new Thread("ButtonBarPerformer")
         {
            public void run ()
            {
               performButton( e.getSource() );
            }
         };
         thread.start();
      }
      else
         performButton( e.getSource() );
   }

   /**
	 * Sets a <code>ButtonBarListener</code> for this bar that will 
    * handle button actions according to the needs of the user application.
    * There can be only one such listener for a single bar.
	 * 
	 * @param listener  the <code>ButtonBarListener</code> to activate
	 */
	public void setButtonBarListener( ButtonBarListener listener )
	{
		this.listener = listener;
	}

   /**
    * @return Returns the help button or <b>null</b> if undefined.
    */
   public JButton getHelpButton()
   {
      return helpButton;
   }

	/**
	 * @return Returns the cancelButton or <b>null</b> if undefined.
	 */
	public JButton getCancelButton()
	{
		return cancelButton;
	}

	/**
	 * @return Returns the okButton or <b>null</b> if undefined.
	 */
	public JButton getOkButton()
	{
		return okButton;
	}

	/**
	 * Returns <code>true</code> if the "Cancel" button was pressed by the
    * last user action on this button bar.
	 * 
	 * @return
	 */
	public boolean isCancelPressed()
	{
		return cancelPressed;
	}

	/**
	 * Returns <code>true</code> if the "Ok" button was pressed by the
    * last user action on this button bar.
	 * 
	 * @return
	 */
	public boolean isOkPressed()
	{
		return okPressed;
	}
}
